#
# Copyright 2019 AXA Group Operations S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging


class ParsrOutputInterpreter(object):
    """Functions to interpret Parsr's resultant JSON file, enabling
    access to the underlying document content
    """

    def __init__(self, object=None):
        """Constructor for the class

        - object: the Parsr JSON file to be loaded
        """
        logging.basicConfig(level=logging.DEBUG,
                            format='%(name)s - %(levelname)s - %(message)s')
        self.object = None
        if object is not None:
            self.load_object(object)

    def __get_text_types(self):
        """Internal function returning the types of text structures
        """
        return ['word', 'line', 'character', 'paragraph', 'heading']

    def __text_objects_none_page(self, txts, page_number_none):
        for page in self.object['pages']:
            for element in page['elements']:
                if element['type'] in self.__get_text_types():
                    txts.append(element)

    def __get_text_objects(self, page_number=None):
        texts = []
        if page_number is not None:
            page = self.get_page(page_number)
            if page is None:
                logging.error(
                    "Cannot get text elements for the requested page; Page {} not found".format(page_number))
                return None
            else:
                for element in page['elements']:
                    if element['type'] in self.__get_text_types():
                        texts.append(element)
        else:
            texts = self.__text_object_none_page(texts, page_number)

        return texts

    def __text_from_text_object(self, text_object: dict) -> str:
        result = ""
        if (text_object['type'] in ['paragraph', 'heading']) or (
                text_object['type'] in ['line']):
            for i in text_object['content']:
                result += self.__text_from_text_object(i)
        elif text_object['type'] in ['word']:
            if isinstance(text_object['content'], list):
                for i in text_object['content']:
                    result += self.__text_from_text_object(i)
            else:
                result += text_object['content']
                result += ' '
        elif text_object['type'] in ['character']:
            result += text_object['content']
        return result

    def load_object(self, object):
        self.object = object

    def get_page(self, page_number: int):
        """Get a particular page in a document

        - page_number: The number of the page to be searched
        """
        for p in self.object['pages']:
            if p['pageNumber'] == page_number:
                return p
        logging.error("Page {} not found".format(page_number))
        return None

    def get_text(self, page_number: int = None) -> str:
        """Get the entire text from a particular page

        - page_number: The page number from which all the text is to be
        extracted
        """
        final_text = ""
        for text_obj in self.__get_text_objects(page_number):
            final_text += self.__text_from_text_object(text_obj)
            final_text += "\n\n"
        return final_text



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['parsr_client']

package_data = \
{'': ['*']}

install_requires = \
['diff_match_patch>=20181111,<20181112',
 'pandas>=1.0.3,<2.0.0',
 'requests>=2.23.0,<3.0.0',
 'semver>=2.9.1,<3.0.0',
 'sxsdiff>=0.3.0,<0.4.0']

setup_kwargs = {
    'name': 'parsr-client',
    'version': '3.2.3',
    'description': 'Python client for Parsr - Transforms PDF, Documents and Images into Enriched Structured Data',
    'long_description': "# Parsr Client\n\nProvides a python interface to the Parsr tool via its API.\nParsr transforms PDF, documents and images into enriched, structured data.\n\nFind out all about Parsr (including download) at [https://github.com/axa-group/Parsr](https://github.com/axa-group/Parsr).\n\n## 1 Installation\n\n```sh\npip install parsr-client\n```\n\n## 2 Usage\n\n_Make sure that the Parsr Server is already running. Let us suppose that the address is `localhost:3001`_\n\n### 2.1 Connect to the Parsr server\n\n```python\nfrom parsr_client import ParsrClient\nparsr = ParsrClient('localhost:3001')\n```\n\n### 2.2 Send the document\n\n```python\nparsr.send_document(\n   file_path='README.pdf',\n   config_path='defaultConfig.json'\n   document_name='The Readme',\n   save_request_id=True)\n```\n\n### 2.4 Retrieve results\n\n1. Get everything as a JSON:\n\n    ```python\n    parsr.get_json()\n    ```\n\n2. As Markdown:\n\n    ```python\n    parsr.get_markdown()\n    ```\n\n3. As text:\n\n    ```python\n    parsr.get_text()\n    ```\n\n4. Get the first table on the first page:\n\n    ```python\n    parsr.get_table(\n        page=1,\n        table=1,\n    )\n    ```\n\n5. Get all the versions of the document:\n\n    ```python\n    parsr.get_revisions('The Readme')\n    ```\n\n6. Get pretty diffs between each successive pair of a document's revisions:\n\n    ```python\n    parsr.compare_revisions('The Readme', pretty_html=True)\n    ```\n\n## 3 Interpreting the whole JSON output locally\n\nThe supplied `ParsrOutputInterpreter` class can be used to interpret the downloaded JSON output and generate higher level structures like the text body.\n\nHere's an example to generate text body on the first page from the above example.\n\n``` python\nfrom parsr_client import ParsrOutputInterpreter\n\nparsr_interpreter = ParsrOutputInterpreter(\n    parsr.get_json()\n)\n\nt = parsr_interpreter.get_text(\n    page_number=1\n)\nprint(t)\n```\n",
    'author': 'AXA Group Operations S.A.',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://par.sr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#
# Copyright 2020 AXA Group Operations S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from glob import glob
from itertools import chain
from io import StringIO
from os import path
from os import chdir
from json import loads
from time import sleep
from ast import literal_eval
from pandas import read_csv
from pandas import notnull
from requests import post
from requests import get
from semver import VersionInfo

import diff_match_patch
from sxsdiff import DiffCalculator
from sxsdiff.generators.github import GitHubStyledGenerator


class ParsrClient():
    """The Parsr client class with all the necessary functions to connect to,
    send and receive data to/from the Parsr server using the API.
    """

    def __init__(self, server: str, revision_history: dict = {}):
        """Constructor for the class.

        - server: The address of the server. For example: 'localhost:3001'
        - revision_history: The revision history of a client, if a previous
        state is to be restored.
        """
        self.set_server(server)
        self.set_revision_history(revision_history)
        self.set_current_request_id("")

    def __supported_input_files(self) -> list:
        """Returns the file extensions supported as input files for Parsr
        """
        return ['*.pdf', '*.jpg', '*.jpeg', '*.png', '*.tiff',
                '*.tif', '*.docx', '*.xml', '*.eml' '*.json']

    def set_server(self, server: str):
        """Setter for the Parsr server's address
        """
        self.server = server

    def set_revision_history(self, revision_history: dict):
        """Set a previous revision history for the client
        """
        self.revision_history = revision_history

    def set_current_request_id(self, request_id: str):
        """Set the current request_id for the client
        """
        self.request_id = request_id

    def send_document(
            self,
            file_path: str,
            config_path: str,
            server: str = "",
            document_name: str = None,
            revision: str = 'major',
            wait_till_finished: bool = False,
            refresh_period: int = 2,
            save_request_id: bool = False,
            silent: bool = True) -> dict:
        """Send the document to the Parsr server

        - file: The address of the file to be sent to the server
        - config: The configuration file to be sent with the request
        - document_name: The name of the document
        - revision: Type of revision (major or minor) of the document
        - wait_till_finished: Should the sending be blocking or not
        - refresh_period: How often the server is to be polled to see
        if the job has finished
        - save_request_id: If set to true, will remember the current
        job and it won't have to be specified again
        - silent: If set to false, each polling request will be logged.
        """
        if server == "":
            if self.server == "":
                raise Exception('No server address provided')
            else:
                server = self.server
                packet = {
                    'file': (
                        file_path,
                        open(file_path, 'rb'),
                        'application/pdf'),
                    'config': (config_path,
                               open(config_path, 'rb'),
                               'application/json'),
                }
                r = post(
                    'http://' + server + '/api/v1/document',
                    files=packet)
                jobId = r.text
        if not document_name:
            document_name = path.splitext(path.basename(file_path))[0]
        if document_name not in self.revision_history:
            self.revision_history[document_name] = {
                str(VersionInfo.parse('1.0.0')): jobId
            }
        else:
            latest_revision = max(VersionInfo.parse(i) for i in list(
                self.revision_history[document_name].keys()))
            if revision == 'major':
                new_revision = latest_revision.bump_major()
            elif revision == 'minor':
                new_revision = latest_revision.bump_minor()
                self.revision_history[document_name][str(new_revision)] = jobId
        if save_request_id:
            self.set_current_request_id(jobId)
        if not wait_till_finished:
            return {
                'file': file_path,
                'config': config_path,
                'status_code': r.status_code,
                'server_response': r.text}
        else:
            print('> Polling server for the job {}...'.format(jobId))
            server_status_response = self.get_status(jobId)['server_response']
            while ('progress-percentage' in server_status_response):
                if not silent:
                    print(
                        '>> Progress percentage: {}'.format(
                            server_status_response['progress-percentage']))
                    sleep(refresh_period)
                    server_status_response = self.get_status(jobId)[
                        'server_response']
                    print('>> Job done!')
            return {
                'file': file,
                'config': config,
                'status_code': r.status_code,
                'server_response': r.text}

    def get_request_id(self, document_name: str, revision: str) -> str:
        """Gets the request ID from a document name and a revision ID

        - document_name: The name of the document
        - revision: The ID of the revision being queried
        """
        if document_name in self.revision_history:
            if revision in self.revision_history[document_name].keys():
                return self.revision_history[document_name][revision]
            else:
                print(
                    'Revision {} not found for document {}'.format(
                        revision, document_name))
        else:
            print('Document name {} not found'.format(document_name))
        return ""

    def get_revisions(self, document_name: str) -> list:
        """Get a list of all the revisions given a document name

        - document_name: The name of the document
        """
        if document_name in self.revision_history:
            return list(self.revision_history[document_name].keys())
        else:
            return []

    def get_document_name_from_request_id(self, request_id: str) -> str:
        """Get the name of a document knowing only its request ID

        - request_id: the request id for which the document is to be searched
        """
        for document_name in list(self.revision_history.keys()):
            if request_id in [self.revision_history[document_name][i]
                              for i in list(
                                      self.revision_history[document_name]
                              )]:
                return document_name
        return ""

    def compare_revisions(
            self,
            document_name: str,
            revisions: list = [],
            pretty_html: bool = False) -> list:
        """Compare two documents and return the diff between them

        - document_name: the name of the document
        - revisions: the list of revisions to be taken into consideration.
        When nothing is supplied, all the revisions of the document are
        compared
        - pretty_html: if a general diff is to be returned or should a pretty
        html document is to be generated instead
        """
        diffs = []
        if len(revisions) == 0:
            revisions = self.get_revisions(document_name)
            request_ids = [
                self.get_request_id(
                    document_name,
                    i) for i in revisions]
        for i in range(0, len(request_ids) - 1):
            request_id1 = request_ids[i]
            request_id2 = request_ids[i + 1]
            md1 = self.get_markdown(request_id1)
            md2 = self.get_markdown(request_id2)

            if pretty_html:
                sxsdiff_result = DiffCalculator().run(md1, md2)
                html_store = StringIO()
                GitHubStyledGenerator(file=html_store).run(sxsdiff_result)
                html_diff = html_store.getvalue()
                diffs.append(html_diff)
            else:
                dmp = diff_match_patch.diff_match_patch()
                diff = dmp.diff_main(md1, md2)
                dmp.diff_cleanupSemantic(diff)
                diffs.append(diff)
        return diffs

    def send_documents_folder(
            self,
            folder: str,
            config: str,
            server: str = "") -> list:
        """Send all the files inside a folder

        - folder: The name of the folder to be sent to Parsr
        - config: The address to the configuration file to be used
        for treating the documents
        """
        if server == "":
            if self.server == "":
                raise Exception('No server address provided')
            else:
                server = self.server
                responses = []
                chdir(folder)
                files = [glob(e) for e in self.__supported_input_files()]
                files_flat = list(chain.from_iterable(files))
        for file in files_flat:
            packet = {
                'file': (file, open(file, 'rb'), 'application/pdf'),
                'config': (config, open(config, 'rb'), 'application/json'),
            }
            r = post(
                'http://' +
                server +
                '/api/v1/document',
                files=packet)
            responses.append({'file': file,
                              'config': config,
                              'status_code': r.status_code,
                              'server_response': r.text})
        return responses

    def get_status(self, request_id: str = "", server: str = ""):
        """Get the status of a partcular request using its ID

        - request_id: The ID of the request to be queried with the server
        - server: The server address where the query is to be made
        """
        if server == "":
            if self.server == "":
                raise Exception('No server address provided')
            else:
                server = self.server
        if request_id == "":
            if self.request_id == "":
                raise Exception('No request ID provided')
            else:
                request_id = self.request_id
        if self.server == "":
            raise Exception('No server address provided')
        r = get(
            'http://{}/api/v1/queue/{}'.format(server, request_id))
        return {
            'request_id': request_id,
            'server_response': loads(
                r.text)}

    def get_json(self, request_id: str = "", server: str = ""):
        """Fetch the Parsr's output JSON file (result) given a particular
        request

        - request_id: The ID of the request to be queried with the server
        - server: The server from which the JSON is to be fetched
        """
        if server == "":
            if self.server == "":
                raise Exception('No server address provided')
            else:
                server = self.server
        if request_id == "":
            if self.request_id == "":
                raise Exception('No request ID provided')
            else:
                request_id = self.request_id
                r = get(
                    'http://{}/api/v1/json/{}'.format(server, request_id))
        if r.text != "":
            return r.json()
        else:
            return {'request_id': request_id, 'server_response': r.json()}

    def get_markdown(self, request_id: str = "", server: str = ""):
        """Fetch the Parsr's output Markdown file (result) given a particular
        request

        - request_id: The ID of the request to be queried with the server
        - server: The server from which the result is to be fetched
        """
        if server == "":
            if self.server == "":
                raise Exception('No server address provided')
            else:
                server = self.server
        if request_id == "":
            if self.request_id == "":
                raise Exception('No request ID provided')
            else:
                request_id = self.request_id
                r = get(
                    'http://{}/api/v1/markdown/{}'.format(server, request_id))
        if r.text != "":
            return r.text
        else:
            return {'request_id': request_id, 'server_response': r.text}

    def get_text(self, request_id: str = "", server: str = ""):
        """Fetch the Parsr's output Text file (result) given a particular
        request

        - request_id: The ID of the request to be queried with the server
        - server: The server from which the result is to be fetched
        """
        if server == "":
            if self.server == "":
                raise Exception('No server address provided')
            else:
                server = self.server
        if request_id == "":
            if self.request_id == "":
                raise Exception('No request ID provided')
            else:
                request_id = self.request_id
                r = get(
                    'http://{}/api/v1/text/{}'.format(server, request_id))
        if r.text != "":
            return r.text
        else:
            return {'request_id': request_id, 'server_response': r.text}

    def get_tables_info(self, request_id: str = ""):
        """Fetch the list and location of tables detected inside a particular
        document.
        """
        return [(table.rsplit('/')[-2], table.rsplit('/')[-1])
                for table in literal_eval(
                        self.get_table(request_id=request_id).columns[0])]

    def get_table(
            self,
            request_id: str = "",
            page=None,
            table=None,
            seperator=";",
            server: str = "",
            column_names: list = None):
        """Get a particular table from a processed document.

        - request_id: The request to be queried to get a document.
        - page: The page number on which the queried table exists.
        - table: The table number to be fetched.
        - seperator: The seperator to be used between table cells (default ';')
        - server: The server address which is to be queried.
        - column_names: The headings of the table searched (column titles)
        """
        if server == "":
            if self.server == "":
                raise Exception('No server address provided')
            else:
                server = self.server
        if request_id == "":
            if self.request_id == "":
                raise Exception('No request ID provided')
            else:
                request_id = self.request_id
        if page is None and table is None:
            r = get('http://{}/api/v1/csv/{}'.format(server, request_id))
        else:
            r = get('http://{}/api/v1/csv/{}/{}/{}'.format(server,
                                                           request_id,
                                                           page,
                                                           table))
        if r.text != "":
            try:
                df = read_csv(
                    StringIO(
                        r.text),
                    sep=seperator,
                    names=column_names)
                df.loc[:, ~df.columns.str.match('Unnamed')]
                df = df.where((notnull(df)), " ")
                return df
            except Exception:
                return r.text
        else:
            return r.text
